/* This file is part of Vcsync
   Copyright (C) 2014, 2019 Sergey Poznyakoff
  
   Vcsync is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   Vcsync is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with Vcsync.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "vcsync.h"
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>

int
sigvsetact(int sigc, int *sigv, struct sigaction *sa)
{
	int i;
	
	for (i = 0; i < sigc; i++) {
		if (sigaction(sigv[i], &sa[i], NULL))
			return i+1;
	}
	return 0;
}

int
sigvset(void (*handler)(int), int sigc, int *sigv, struct sigaction *retsa)
{
	int i;
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	for (i = 0; i < sigc; i++) {
		sa.sa_handler = handler;
		
		if (sigaction(sigv[i], &sa, retsa ? &retsa[i] : NULL)) {
			if (retsa) {
				int ec = errno;
				sigvsetact(i, sigv, retsa);
				errno = ec;
			}
			return -1;
		}
	}
	return 0;
}

void
sigsetup(void (*sf) (int))
{
	static int sigv[] = { SIGALRM, SIGCHLD };
	sigvset(sf, sizeof(sigv)/sizeof(sigv[0]), sigv, NULL);
}


/* Close all file descriptors, except those in fdset. */
static void
close_fds(fd_set *fdset)
{
        int i;

        for (i = FD_SETSIZE-1; i >= 0; i--) {
                if (FD_ISSET(i, fdset))
                        continue;
                close(i);
        }
}

/* Create a redirector, i.e. a process that reads its standard input and
   sends it to syslog with the priority PRIO.  Syslog records will be marked
   with TAG.  Upon successful return, stores the PID of the created process
   in RETURN_PID and returns the file descriptor sutable for redirecting
   output to it. */
static int
open_redirector(const char *tag, int prio, pid_t *return_pid)
{
	int p[2];
	FILE *fp;
	char buf[512];
	pid_t pid;
	fd_set fdset;

	if (pipe(p)) {
		diag(LOG_ERR,
		     "cannot start redirector for %s, pipe failed: %s",
		     tag, strerror(errno));
		return -1;
	}
	switch (pid = fork()) {
	case 0:
		/* Redirector process */
		FD_ZERO(&fdset);
		FD_SET(p[0], &fdset);
		if (log_to_stderr >= 0 || cfg_facility <= 0)
			FD_SET(2, &fdset);
		close_fds(&fdset);
		sigsetup(SIG_DFL);

		fp = fdopen(p[0], "r");
		if (fp == NULL)
			_exit(1);
		if (cfg_facility > 0) 
			openlog(tag, LOG_PID, cfg_facility);

		while (fgets(buf, sizeof(buf), fp) > 0) {
			int len = strlen(buf);
			if (len && buf[len-1] == '\n')
				buf[len-1] = 0;
			diag(prio, "%s", buf);
		}
		_exit(0);
      
	case -1:
		diag(LOG_CRIT,
		     "cannot run redirector `%s': fork failed: %s",
		     tag, strerror(errno));
		return -1;

	default:
		debug(2, ("redirector for %s started, pid=%lu",
			  tag, (unsigned long) pid));
		close(p[0]);

		*return_pid = pid;
		return p[1];
	}
}

#define REDIR_ERR 0
#define REDIR_OUT 1

#define PID_MASTER 0
#define PID_ERR    1
#define PID_OUT    2

static char *cmdname;
static pid_t pidtab[3] = { -1, -1, -1 };
static int npid;
static int status;

void
sigmain(int sig)
{
	pid_t pid;
	int s;
	
	switch (sig) {
	case SIGCHLD:
		while ((pid = waitpid(-1, &s, WNOHANG)) > 0) {
			int i;
			
			for (i = 0; i < 3; i++)
				if (pidtab[i] == pid) {
					if (i == PID_MASTER)
						status = s;
					pidtab[i] = -1;
					npid--;
				}
		}
		break;

	case SIGALRM:
		diag(LOG_CRIT, "timed out");
		exit(EX_FAIL);
	}
}

void
kill_redirectors()
{
	if (pidtab[PID_OUT] > 1)
		kill(pidtab[PID_OUT], SIGKILL);
	if (pidtab[PID_ERR] > 1)
		kill(pidtab[PID_ERR], SIGKILL);
}

int
runcmd(char **argv, int *pfd)
{
	fd_set fdset;
	int redir_fd[2] = { -1, -1 };
	int p[2];

	if (debug_level >= 1) {
		int i;
		struct grecs_txtacc *acc = grecs_txtacc_create();
		
		grecs_txtacc_grow_string(acc, argv[0]);
		for (i = 1; argv[i]; i++) {
			grecs_txtacc_grow_char(acc, ' ');
			grecs_txtacc_grow_string_escape(acc, argv[i]);
		}
		grecs_txtacc_grow_char(acc, 0);

		debugprt("starting %s", grecs_txtacc_finish(acc, 0));
		grecs_txtacc_free(acc);
	}

	if (dry_run)
		return 0;
	
	npid = 1;
	redir_fd[REDIR_ERR] = open_redirector(argv[0], LOG_ERR,
					      &pidtab[PID_ERR]);

	if (pfd)
		pipe(p);
	else {
		npid++;
		redir_fd[REDIR_OUT] = open_redirector(argv[0], LOG_INFO,
						      &pidtab[PID_OUT]);
	}
	
	npid++;
	pidtab[PID_MASTER] = fork();
	if (pidtab[PID_MASTER] == -1) {
		diag(LOG_ERR, "fork: %s", strerror(errno));
		close(redir_fd[REDIR_OUT]);
		close(redir_fd[REDIR_ERR]);
		if (pfd) {
			close(p[0]);
			close(p[1]);
		}
		kill_redirectors();
		exit(EX_FAIL);
	}

	if (pidtab[PID_MASTER] == 0) {
		/* child */
		int fd;

		FD_ZERO(&fdset);
		fd = open("/dev/null", O_RDONLY);
		if (fd != -1) {
			if (fd != 0)
				dup2(fd, 0);
			FD_SET(0, &fdset);
		}
		if (pfd) {
			if (p[1] != 1 &&
			    dup2(p[1], 1) == -1) {
				diag(LOG_ERR, "dup2: %s", strerror(errno));
				_exit(127);
			}
			FD_SET(1, &fdset);
		} else if (redir_fd[REDIR_OUT] != -1) {
			if (redir_fd[REDIR_OUT] != 1 &&
			    dup2(redir_fd[REDIR_OUT], 1) == -1) {
				diag(LOG_ERR, "dup2: %s", strerror(errno));
				_exit(127);
			}
			FD_SET(1, &fdset);
		}
		if (redir_fd[REDIR_ERR] != -1) {
			if (redir_fd[REDIR_ERR] != 2 &&
			    dup2(redir_fd[REDIR_ERR], 2) == -1) {
				diag(LOG_ERR, "dup2: %s", strerror(errno));
				_exit(127);
			}
			FD_SET(2, &fdset);
		}
		close_fds(&fdset);
		sigsetup(SIG_DFL);
		execv(argv[0], argv);
		diag(LOG_CRIT, "execve(%s): %s", argv[0], strerror(errno));
		kill_redirectors();
		_exit(127);
	}

	/* Master */
	cmdname = estrdup(argv[0]);
	
	close(redir_fd[REDIR_ERR]);

	if (pfd) {
		*pfd = p[0];
		close(p[1]);
		return 0;
	}
	
	close(redir_fd[REDIR_OUT]);

	return runcmd_wait();
}

int
runcmd_wait()
{
	alarm(config->timeout);
	while (npid)
		pause();
	alarm(0);
	
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) == 0) {
			debug(1, ("%s terminated successfully", cmdname));
			free(cmdname);
			return 0;
		} else
			diag(LOG_NOTICE,
			     "%s exited with status %d",
			     cmdname, WEXITSTATUS(status));
	} else if (WIFSIGNALED(status))
		diag(LOG_NOTICE,
		     "%s terminated on signal %d",
		     cmdname, WTERMSIG(status));
	else if (WIFSTOPPED(status))
		diag(LOG_NOTICE, "%s stopped", cmdname);

#ifdef WCOREDUMP
	if (WCOREDUMP(status))
		diag(LOG_NOTICE, "%s dumped core", cmdname);
#endif
	free(cmdname);
	return 1;
}

